

// Arrays.

// Lines labelled ``bodge'' are clumsy fixes to support the Fortran
// run-time library.


//#include <stdlib.h>
//#include <iostream.h>
//#include <stdarg.h>
extern "C" {
  int printf(char*, ...) ;
}
//extern "C" {
//  #include "mpi.h"
//}  // debug
#include "Seq.h"
#include "Procs.h"
#include "Range.h"
#include "Array.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Stride': Generalises the idea of an array index stride.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline int abs(int x) {return x > 0 ? x : - x ; }

Stride :: Stride() : base_v(0), stride_v(0), norm_h(1), stride_h(0) {}

Stride :: Stride(const Stride& s) : base_v(s.base_v), stride_v(s.stride_v),
                                    norm_h(s.norm_h), stride_h(s.stride_h),
                                    stodge(s.stodge)  // bodge
                                    {}

int Stride :: init(Range* x, const int dataLen) {

  // The representation of the stride structure is independent
  // of the `x' coordinate!

  const int stride    = x->stride ;
  const int layerSize = x->layerSize ;
  const int s_lo      = x->base   ;
  const int s_hi      = x->base + (x->extent - 1) * stride ;

  int blockWidth ;
  int top_v ;

  if(stride == 0) AdlibError("Stride :: init: stride is zero") ;

  if(stride > 0) {
    blockWidth = (x->blockSize + stride - 1) / stride ;
    top_v      = s_hi / layerSize ;

    stride_v = blockWidth * dataLen ;
    base_v   = s_lo / layerSize ;
    stride_h = dataLen ;
    norm_h   = stride ;
  }
  else {
    blockWidth = (x->blockSize - stride - 1) / (-stride) ;
    top_v      = s_lo / layerSize ;

    stride_v = blockWidth * dataLen ;
    base_v   = s_hi / layerSize ;
    stride_h = dataLen ;
    norm_h   = -stride ;
  }

  stodge = !x->distributed() ;  // bodge

  return (top_v - base_v + 1) * blockWidth * dataLen ;
      // Allocates a layer even if size = 0.  Not perfect.
}


int Stride :: offset(Subscript* i) {

  // Offset generated by subscript `i'

  if(stodge && i->distributed()) return stride_h * i->bodge ;  // bodge
  
  return stride_v * (i->s_v - base_v) + stride_h * (i->s_h / norm_h) ;
}

int Stride :: step(Range* x) {

  // Address increment between adjacent values in range (when
  // this increment is ``simple'').

  if(stodge && x->distributed) return stride_h ;  // bodge

  return stride_h * (x->stride / norm_h) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Frame': Representation of a general array entity.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Frame :: Frame(int _typeLen, ProcessGroup* _procGrp, const int _rank) :
             typeLen(_typeLen), rank(_rank) {

  // Create a new array.

  // `_procGrp' is copied into `Frame' structure.
  // `_rngs' elements are installed in the `Frame' structure (without copying).
  // `strs' is allocated and initialised.
  // `data' is allocated.

  procGrp = new ProcessGroup(*_procGrp) ;

  rngs = new Range*  [rank] ;
  strs = new Stride* [rank] ;
}


Frame :: Frame(Frame* parent, Subscript** subs, const int _rank) :
             typeLen(parent->typeLen), rank(_rank) {

  // Create an array section by direct subscripting.

  // `subs' is null terminated ``string'' of pointers.

  procGrp = new ProcessGroup(*(parent->procGrp)) ;
  procGrp->activeSection() ;

  rngs = new Range*  [rank] ;
  strs = new Stride* [rank] ;

  data = parent->data ;

  int i = 0 ;
  for(int j = 0 ; j < parent->rank ; j++) {
    Subscript* x = subs [j] ;

    if(!x)
      AdlibError("Frame :: Frame: not enough section subscripts for array") ;

    // Following accesses to `Range' fields seem fairly gratuitous
    if(x->isRange) {
      if(x->dim)
	if(!procGrp->isActive(x->dim))
	  AdlibError(
	      "Frame :: Frame: range dimension not active in process group") ;
      if(i >= rank)
	AdlibError("Frame :: Frame: too many range-valued subscripts") ;
      rngs [i] = (Range*) x ;
      strs [i] = parent->strs [j] ;
      i++ ;
    }
    else {
      data += ((parent->strs) [j])->offset(x) ;
      if(x->distributed())
	procGrp->deactivate(x) ;
    }
  }
  if(rank < i)
    AdlibError("Frame :: Frame: "
               "not enough range-valued subscripts in section") ;

  if(subs [parent->rank])
    AdlibError("Frame :: Frame: too many section subscripts for array") ;
}

//Frame :: Frame(Vector* focus,
//               Embed* _embs, int _rank, Range* _rngs, Stride* _strs) :
//    embs(_embs), rank(_rank), rngs(_rngs), strs(_strs) {
//
//  // Create an array section from a focus.
//
//  // `_rngs' and `_strs' should be preallocated vectors.  
//  // `_rngs' should contain the range list of the section.
//
//  // `_strs' is initalised, and both vectors are installed in the
//  // `Frame' structure.
//
//  size = 0 ;
//  data = focus->data ;
//}

Frame :: ~Frame() {
  delete strs ;
  delete rngs ;
  delete procGrp ;
}


char* Frame :: subs(Subscript** subs) {
  char* address = data ;
 
  for (int i = 0; i < rank ; i++)
    address += (strs [i])->offset(subs [i]) ;

  return address ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Data': Representation of an array variable.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Data :: Data(const int _typeLen, ProcessGroup* _procGrp,
             const int _rank, Range** _rngs) :
                 Frame(_typeLen, _procGrp, _rank) {

  // Create a new array.

  // `_procGrp' is copied into `Frame' structure.
  //  `_rngs' elements are installed in the `Frame' structure (without copying).
  // `strs' is allocated and initialised.
  // `data' is allocated.

  int size = typeLen ;
  for(int i = 0 ; i < rank ; i++) {
    Range* x = _rngs [i] ;
    if(x->dim)
      if(!procGrp->isActive(x->dim))
	AdlibError(
            "Data :: Data: range dimension not active in process group") ;
      
    rngs [i] = x ;

    Stride* str = new Stride ;
    size = str->init(x, size) ;
    strs [i] = str ;
  }

  data = new char [size] ;
}

Data :: ~Data() {
  delete data ;
  for(int i = 0 ; i < rank ; i++)
    delete strs [i] ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Promotion': an extension of an array, with added ranges      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Promotion :: Promotion(const int _rank, Range** _rngs,
                       Frame* parent, int* injection) :
                 Frame(parent->typeLen, parent->procGrp, _rank) {

  // General promotion---add any number of ranges, with any injection
  // of the old ranges into the new.

  int i ;

  data = parent->data ;

  for(i = 0 ; i < rank ; i++) {
    rngs [i] = _rngs [i] ;
    strs [i] = NULL ;
  }

  for(int j = 0 ; j < parent->rank ; j++)
    strs [injection [j] - 1] = new Stride(*(parent->strs [j])) ;

  for(i = 0 ; i < rank ; i++)
    if(!strs [i])
      strs [i] = new Stride ;
}

Promotion :: Promotion(Frame* parent, int dim, Range* x) :
                 Frame(parent->typeLen, parent->procGrp, parent->rank + 1) {

  // Elementary promotion---analogue of Fortran 90 `spread'.

  data = parent->data ;

  int j = 0 ;
  for(int i = 0 ; i < rank ; i++)
    if(i == dim - 1) {
      strs [i] = new Stride ;
      rngs [i] = x ;
    }
    else {
      strs [i] = new Stride(*(parent->strs [j])) ;
      rngs [i] = parent->rngs [j] ;
      j++ ;
    }
}

Promotion :: Promotion(const int _typeLen, ProcessGroup* _procGrp,
		       const int _rank, Range** _rngs, char* _data) :
			   Frame(_typeLen, _procGrp, _rank) {

  // Promotion of scalar value to array value.

  data = _data ;

  for(int i = 0 ; i < rank ; i++) {
    strs [i] = new Stride ;
    rngs [i] = _rngs [i] ;
  }
}

Promotion :: ~Promotion() {
  for(int i = 0 ; i < rank ; i++)
    delete strs [i] ;
}



//// `Vector' is a lightweight representation of an array section,
//// used for traversing arrays in distributed loops.
//
//
//void Vector :: first(int d, Range* x) {
//  if(x->loopTest) {
//    // Push old values of `data' and `incr' onto ``stack''
//    base [d] = data ;
//    step [d] = incr ;
//
//    incr = ((parent->strs) [d]).step(x) ;
//    data = ((parent->strs) [d]).address(base [d], x) ;
//  }
//}
//
//void Vector :: next(int d, Range* x) {
//  data += incr ;
//  if(!x->simpleTest)
//    if(x->loopTest)
//      data = ((parent->strs) [d]).address(base [d], x) ;
//  if(!x->loopTest) {
//    // Pop old values of `data' and `incr' off ``stack''
//    incr = step [d] ;
//    data = base [d] ;
//  }
//}


// --------------------
// Things to do
//
